/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 David N. Welton.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        David N. Welton (http://www.dedasys.com/davidw/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "CacheOut" and "David N. Welton" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    davidw@apache.org.
 *
 * 5. Products derived from this software may not be called "CacheOut",
 *    nor may "CacheOut" appear in their name, without prior written
 *    permission of the author(s).
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * ==================================================================== */

#include <tcl.h>
#include <errno.h>

struct ChanInfo {
    Tcl_Obj *obuff;
    Tcl_Channel old;
    Tcl_Channel self;
    int open;
};

static int inputproc(ClientData instancedata, char *buf, int toRead, int *errorCodePtr)
{
    return EINVAL;
}

static int outputproc(ClientData instancedata, char *buf, int toWrite, int *errorCodePtr)
{
    struct ChanInfo *ChanInfo = instancedata;

    Tcl_AppendObjToObj(ChanInfo->obuff, Tcl_NewByteArrayObj(buf, toWrite));
    return toWrite;
}

static int closeproc(ClientData instancedata, Tcl_Interp *interp)
{
    return 0;
}

static int setoptionproc(ClientData instancedata, Tcl_Interp *interp, char *optionname, char *value)
{
    return TCL_OK;
}

static void watchproc(ClientData instancedata, int mask)
{
    return;
}

static int gethandleproc(ClientData instancedata, int direction, ClientData *handlePtr)
{
    return TCL_ERROR;
}

static Tcl_ChannelType CacheChan = {
    "cache_channel",
    NULL,
    closeproc,
    inputproc,
    outputproc,
    NULL,
    setoptionproc,
    NULL,
    watchproc,
    gethandleproc,
    NULL
};

static int CloseCache(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    struct ChanInfo *ChanInfo = clientData;
    if (ChanInfo->open != 1)
    {
	Tcl_AddErrorInfo(interp, "Cache not open");
	return TCL_ERROR;
    }
    Tcl_UnstackChannel(interp, ChanInfo->self);
    Tcl_SetObjResult(interp, Tcl_DuplicateObj(ChanInfo->obuff));
    Tcl_SetStringObj(ChanInfo->obuff, "", -1);
    ChanInfo->open = 0;
    return TCL_OK;
}

static int OpenCache(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    struct ChanInfo *ChanInfo = clientData;
    if (ChanInfo->open != 0)
    {
	Tcl_AddErrorInfo(interp, "Cache currently open");
	return TCL_ERROR;
    }
    ChanInfo->self = Tcl_StackChannel(interp, &CacheChan, ChanInfo, TCL_WRITABLE,  ChanInfo->old);
    ChanInfo->open = 1;
    return TCL_OK;
}


int Cacheout_Init(Tcl_Interp *interp)
{
    struct ChanInfo *ChanInfo;

    ChanInfo = (struct ChanInfo *)Tcl_Alloc(sizeof(struct ChanInfo));
    ChanInfo->obuff = Tcl_NewObj();
    Tcl_IncrRefCount(ChanInfo->obuff);
    ChanInfo->old = Tcl_GetStdChannel(TCL_STDOUT);
    ChanInfo->open = 0;

    Tcl_CreateObjCommand(interp, "opencache", OpenCache, ChanInfo, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "closecache", CloseCache, ChanInfo, (Tcl_CmdDeleteProc *)NULL);
    Tcl_PkgProvide(interp, "CacheOut", "1.0");

    return TCL_OK;
}
